from pathlib import Path

import numpy as np
import pytest

_IRIS_FILE = "iris.csv"
# maturin build maps the parent directory of the python package to
# "local_dependencies/changeforest". This allows tests to run e.g. on conda-forge.
local_dependencies = Path(__file__).resolve().parents[1] / "local_dependencies"
if local_dependencies.exists():
    _IRIS_PATH = local_dependencies / "changeforest" / "testdata" / _IRIS_FILE
else:
    _IRIS_PATH = Path(__file__).resolve().parents[2] / "testdata" / _IRIS_FILE


@pytest.fixture(scope="module")
def iris_dataset():
    return np.loadtxt(_IRIS_PATH, skiprows=1, delimiter=",", usecols=(0, 1, 2, 3))


@pytest.fixture(scope="module")
def X_test():
    return np.array(
        [
            [0.0, 0.0, 0.0],
            [0.0, 0.0, 0.0],
            [0.0, 0.0, 0.0],
            [0.0, 0.0, 1.0],
            [0.0, 0.0, 1.0],
            [1.0, 1.0, 1.0],
            [1.0, 1.0, 1.0],
            [1.0, 1.0, 1.0],
            [1.0, 1.0, 1.0],
            [1.0, 1.0, 1.0],
        ]
    )


@pytest.fixture(scope="module")
def X_correlated():
    X = np.zeros((100, 2))

    rng = np.random.default_rng(7)

    X[:, 0] = rng.normal(0, 1, 100)
    X[0:50, 1] = rng.normal(0, 1, 50)
    X[50:100, 1] = X[50:100, 0]

    return X



import numpy as np
import pytest

from changeforest import Control, changeforest


@pytest.mark.parametrize(
    "data, segmentation_type, method, kwargs, expected",
    [
        # minimal_relative_segment_length
        ("iris", "bs", "knn", {"minimal_relative_segment_length": 0.05}, [50, 100]),
        ("iris", "bs", "knn", {"minimal_relative_segment_length": 0.4}, [60]),
        # minimal_gain_to_split
        (
            "iris",
            "bs",
            "change_in_mean",
            {"minimal_gain_to_split": 150 * 0.1},
            [50, 100],
        ),
        ("iris", "bs", "change_in_mean", {"minimal_gain_to_split": 150 * 1}, [50]),
        ("iris", "bs", "change_in_mean", {"minimal_gain_to_split": 150 * 10}, []),
        ("iris", "bs", "change_in_mean", {"minimal_gain_to_split": 150 * 10.0}, []),
        (
            "iris",
            "bs",
            "change_in_mean",
            {"minimal_gain_to_split": None},
            [50, 100],
        ),  # log(150) * 4 / 150
        # model_selection_alpha
        ("iris", "bs", "knn", {"model_selection_alpha": 0.001}, []),
        ("iris", "bs", "knn", {"model_selection_alpha": 0.05}, [50, 100]),
        # random_forest_n_estimators
        # This is impressive and unexpected.
        ("iris", "bs", "random_forest", {"random_forest_n_estimators": 1}, [47, 99]),
        ("iris", "bs", "random_forest", {"random_forest_n_estimators": 100}, [50, 100]),
        # Use X_test instead
        ("X_test", "bs", "random_forest", {"random_forest_n_estimators": 1}, []),
        ("X_test", "bs", "random_forest", {"random_forest_n_estimators": 1.0}, []),
        ("X_test", "bs", "random_forest", {"random_forest_n_estimators": 100}, [5]),
        ("X_correlated", "bs", "random_forest", {"random_forest_max_depth": 1}, []),
        ("X_correlated", "bs", "random_forest", {"random_forest_max_depth": 2}, [49]),
        (
            "X_correlated",
            "bs",
            "random_forest",
            {"random_forest_max_features": "sqrt"},
            [49],
        ),
        ("iris", "bs", "random_forest", {"model_selection_n_permutations": 10}, []),
    ],
)
def test_control_model_selection_parameters(
    iris_dataset,
    X_test,
    X_correlated,
    data,
    method,
    segmentation_type,
    kwargs,
    expected,
):
    if data == "iris":
        X = iris_dataset
    elif data == "X_test":
        X = X_test
    else:
        X = X_correlated

    result = changeforest(X, method, segmentation_type, Control(**kwargs))
    np.testing.assert_array_equal(result.split_points(), expected)


@pytest.mark.parametrize(
    "segmentation_type, kwargs, expected_number_of_segments",
    [
        # seeded_segments_alpha
        (
            "sbs",
            {
                "minimal_relative_segment_length": 0.05,
                "seeded_segments_alpha": 1 / np.sqrt(2),
            },
            44,
        ),
        (
            "sbs",
            {"minimal_relative_segment_length": 0.05, "seeded_segments_alpha": 0.5},
            25,
        ),
        ("bs", {}, 0),
        # number_of_wild_segments
        ("wbs", {"number_of_wild_segments": 10}, 10),
        ("wbs", {"number_of_wild_segments": 25}, 25),
    ],
)
def test_control_segmentation_parameters(
    iris_dataset, segmentation_type, kwargs, expected_number_of_segments
):
    result = changeforest(
        iris_dataset, "change_in_mean", segmentation_type, Control(**kwargs)
    )
    # For each split, add evaluation on left / right segment to segments.
    expected_number_of_segments = (
        expected_number_of_segments + 2 * len(result.split_points()) + 1
    )
    assert len(result.segments) == expected_number_of_segments


@pytest.mark.parametrize(
    "key, default_value, another_value",
    [
        ("random_forest_n_estimators", 100, 11),
        ("minimal_relative_segment_length", 0.01, 0.05),
        ("seed", 0, 1),
        ("random_forest_max_features", "default", 1),
        ("random_forest_max_depth", 8, None),
    ],
)
def test_control_defaults(iris_dataset, key, default_value, another_value):
    result = changeforest(iris_dataset, "random_forest", "bs", Control())
    default_result = changeforest(
        iris_dataset, "random_forest", "bs", Control(**{key: default_value})
    )
    another_result = changeforest(
        iris_dataset, "random_forest", "bs", Control(**{key: another_value})
    )

    assert str(result) == str(default_result)
    assert str(result) != str(another_result)



import numpy as np

from .changeforest import BinarySegmentationResult, OptimizerResult

try:
    import matplotlib.pyplot as plt

    MATPLOTLIB_INSTALLED = True
except ImportError:
    MATPLOTLIB_INSTALLED = False


def _plot_two_step_search_result(optimizer_result):
    if not MATPLOTLIB_INSTALLED:
        raise ImportError("The matplotlib package is required for OptimizerResult.plot")

    min_gain = min(
        min(gain_result.gain) for gain_result in optimizer_result.gain_results
    )
    max_gain = max(
        max(gain_result.gain) for gain_result in optimizer_result.gain_results
    )
    delta = max_gain - min_gain
    (ymin_gain, ymax_gain) = (min_gain - 0.05 * delta, max_gain + 0.05 * delta)
    gain_range = (min_gain - 0.1 * delta, max_gain + 0.1 * delta)

    ymin_proba = -0.05
    ymax_proba = 1.05
    proba_range = (-0.1, 1.1)

    fig, axes = plt.subplots(nrows=4, ncols=2)

    for idx in range(4):
        axes[idx, 0].plot(
            range(optimizer_result.start, optimizer_result.stop),
            optimizer_result.gain_results[idx].gain,
            color="k",
        )
        axes[idx, 0].set_ylim(*gain_range)

        axes[idx, 0].vlines(
            optimizer_result.gain_results[idx].guess,
            ymin=ymin_gain,
            ymax=ymax_gain,
            linestyles="dashed",
            color="#4477AA",  # blue
            linewidth=2,
        )
        axes[idx, 0].vlines(
            optimizer_result.start
            + np.nanargmax(optimizer_result.gain_results[idx].gain),
            ymin=ymin_gain,
            ymax=ymax_gain,
            linestyles="dotted",
            color="#EE6677",  # red
            linewidth=2,
        )
        axes[idx, 0].set_ylabel("approx. gain")

        axes[idx, 1].set_ylim(*proba_range)
        axes[idx, 1].vlines(
            optimizer_result.gain_results[idx].guess,
            ymin=ymin_proba,
            ymax=ymax_proba,
            linestyles="dashed",
            color="#4477AA",  # blue
            linewidth=2,
        )
        axes[idx, 1].scatter(
            range(optimizer_result.start, optimizer_result.stop),
            optimizer_result.gain_results[idx].predictions,
            s=2,
            c="k",
        )

        axes[idx, 1].set_ylabel("proba. predictions")
        if idx < 3:
            axes[idx, 0].set_xticklabels([])
            axes[idx, 1].set_xticklabels([])

    axes[-1, 0].set_xlabel("split")
    axes[-1, 1].set_xlabel("t")

    fig.align_ylabels(axes[:, 0])
    fig.align_ylabels(axes[:, 1])
    fig.tight_layout()
    return fig


def _plot_gain_result(gain_result):
    if not MATPLOTLIB_INSTALLED:
        raise ImportError("The matplotlib package is required for OptimizerResult.plot")

    fig, axes = plt.subplots()

    axes.plot(range(gain_result.start, gain_result.stop), gain_result.gain, color="k")
    axes.vlines(
        np.nanmax(gain_result.gain) + gain_result.start,
        linestyles="dotted",
        color="#EE6677",  # red
        linewidth=2,
    )

    axes.set_xlabel("split")
    axes.set_ylabel("gain")

    fig.tight_layout()
    return fig


def _plot_optimizer_result(optimizer_result):
    if len(optimizer_result.gain_results) == 4:
        return _plot_two_step_search_result(optimizer_result)
    else:
        return _plot_gain_result(optimizer_result.gain_results[-1])


def _plot_binary_segmentation_result(binary_segmentation_result, max_depth=5):
    if not MATPLOTLIB_INSTALLED:
        raise ImportError(
            "The matplotlib package is required for BinarySegmentationResult.plot"
        )

    nodes = [binary_segmentation_result]
    gains = []
    splits = []
    found_changepoints = []
    guesses = []
    n = binary_segmentation_result.stop

    for _ in range(0, max_depth):
        new_nodes = []

        if len(nodes) == 0:
            break

        gains.append([])
        found_changepoints.append([])
        splits.append([])
        guesses.append([])

        for node in nodes:
            splits[-1].append(node.start)
            splits[-1].append(node.stop)

            if node.optimizer_result is not None:
                result = node.optimizer_result.gain_results[-1]
                gains[-1].append(np.full(n, np.nan))
                gains[-1][-1][node.start : node.stop] = result.gain  # noqa: E203
                guesses[-1].append(result.guess)

            if node.model_selection_result.is_significant:
                found_changepoints[-1].append(node.best_split)

            if node.left is not None:
                new_nodes.append(node.left)
            if node.right is not None:
                new_nodes.append(node.right)

        nodes = new_nodes

    depth = len(gains)

    fig, axes = plt.subplots(nrows=depth)
    if depth == 1:
        axes = [axes]

    for idx in range(depth):

        for gain in gains[idx]:
            axes[idx].plot(gain, color="black")

        ymin, ymax = axes[idx].get_ylim()
        new_ymin, new_ymax = ymin - 0.05 * (ymax - ymin), ymax + 0.05 * (ymax - ymin)

        axes[idx].vlines(splits[idx], color="black", ymin=new_ymin, ymax=new_ymax)

        axes[idx].vlines(
            guesses[idx],
            color="#4477AA",  # blue
            ymin=ymin,
            ymax=ymax,
            linewidth=2,
            linestyles="dashed",
        )

        axes[idx].vlines(
            found_changepoints[idx],
            color="#EE6677",  # red
            ymin=ymin,
            ymax=ymax,
            linewidth=2,
            linestyles="dotted",
        )

        axes[idx].set_xlim(0, n)
        axes[idx].set_ylim(new_ymin, new_ymax)
        axes[idx].set_ylabel("approx. gain")
        if idx < depth - 1:
            axes[idx].set_xticklabels([])

    axes[-1].set_xlabel("split")

    return fig


OptimizerResult.plot = _plot_optimizer_result
BinarySegmentationResult.plot = _plot_binary_segmentation_result



from .changeforest import changeforest
from .control import Control
from .plotting import BinarySegmentationResult, OptimizerResult

__all__ = ["Control", "changeforest", "BinarySegmentationResult", "OptimizerResult"]



import numpy as np
import pytest

from changeforest import Control, changeforest


@pytest.mark.parametrize("method", ["knn", "change_in_mean", "random_forest"])
@pytest.mark.parametrize("segmentation_type", ["sbs", "wbs", "bs"])
def test_changeforest(iris_dataset, method, segmentation_type):
    result = changeforest(
        iris_dataset,
        method,
        segmentation_type,
        control=Control(minimal_relative_segment_length=0.1),
    )
    np.testing.assert_array_equal(result.split_points(), [50, 100])


def test_changeforest_repr(iris_dataset):
    result = changeforest(iris_dataset, "random_forest", "bs")
    assert (
        result.__repr__()
        == """\
                    best_split max_gain p_value
(0, 150]                    50   96.233   0.005
 ¦--(0, 50]                  2  -14.191       1
 °--(50, 150]              100   52.799   0.005
     ¦--(50, 100]           53     5.44   0.245
     °--(100, 150]         136   -2.398   0.875\
"""
    )
